// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//=============================================================================
// File Name: WeightVector.cc
// modifed by Tianbao Yang (01.07.2013)
// based on Shai Shalev-Shwartz's file in Pegasos code http://www.cs.huji.ac.il/~shais/code/
// implements the custormized methods and functionalities of the WeightVector class
//=============================================================================

#include "WeightVector.h"

void WeightVector::add(simple_sparse_vector& x, double s) {
  double pred = 0.0, norm_x = 0.0;
  for( simple_sparse_vector_iterator it = x.my_vec.begin(); 
       it != x.my_vec.end(); it++) {
    double val = (*it).second * s;
    norm_x += val*val;
    pred += 2.0*my_v[(*it).first]*val;
    my_1norm -=fabs(my_a*my_v[(*it).first]); // add to handle the l1 regularization
    my_v[(*it).first] += (val/my_a);
    my_1norm +=fabs(my_a*my_v[(*it).first]); // add to handle the l1 regularization
  }
  my_snorm += norm_x + my_a*pred;
}

void WeightVector::add(dense_vector& x, double s) {
	  my_snorm = 0.0; my_1norm=0.0;
	  for (uint i=0; i<d; ++i) {
	    my_v[i] *= my_a;
	    my_v[i] += (x[i] * s);
	    my_snorm += my_v[i]*my_v[i];
	    my_1norm += fabs(my_v[i]);
	  }
	  my_a = 1.0;
}


// this += s*x
void WeightVector::add(WeightVector& x, double s) {
  my_snorm = 0.0; my_1norm=0.0;
  for (uint i=0; i<d; ++i) {
    my_v[i] *= my_a;
    my_v[i] += (x[i] * s);
    my_snorm += my_v[i]*my_v[i];
    my_1norm += fabs(my_v[i]);
  }
  my_a = 1.0;
    
}

template void WeightVector::add(std::vector<float>& x, double s);
template void WeightVector::add(std::vector<double>& x, double s);

template<typename T>
void WeightVector::add(std::vector<T>& x, double s) {
  my_snorm = 0.0; my_1norm=0.0;
  for (uint i=0; i<d; ++i) {
    my_v[i] *= my_a;
    my_v[i] += (x[i] * s);
    my_snorm += my_v[i]*my_v[i];
    my_1norm +=fabs(my_v[i]);
  }
  my_a = 1.0;  
}




void WeightVector::minus(simple_sparse_vector& x, double s) {
  double pred = 0.0, norm_x = 0.0;
  for( simple_sparse_vector_iterator it = x.my_vec.begin(); 
       it != x.my_vec.end(); it++) {
    double val = (*it).second * s;
    norm_x += val*val;
    pred += 2.0*my_v[(*it).first]*val;
    my_1norm -=fabs(my_a*my_v[(*it).first]); // add to handle the l1 regularization
    my_v[(*it).first] -= (val/my_a);
    my_1norm +=fabs(my_a*my_v[(*it).first]); // add to handle the l1 regularization
  }
  my_snorm += norm_x - my_a*pred;
}


// this += s*x
void WeightVector::minus(dense_vector& x, double s) {
  my_snorm = 0.0; my_1norm=0.0;
  for (uint i=0; i<d; ++i) {
    my_v[i] *= my_a;
    my_v[i] -= (x[i] * s);
    my_snorm += my_v[i]*my_v[i];
    my_1norm +=fabs(my_v[i]);
  }
  my_a = 1.0;  
}

template void WeightVector::minus(std::vector<float>& x, double s);
template void WeightVector::minus(std::vector<double>& x, double s);
template<typename T>
void WeightVector::minus(std::vector<T>& x, double s) {
  my_snorm = 0.0; my_1norm=0.0;
  for (uint i=0; i<d; ++i) {
    my_v[i] *= my_a;
    my_v[i] -= (x[i] * s);
    my_snorm += my_v[i]*my_v[i];
    my_1norm +=fabs(my_v[i]);
  }
  my_a = 1.0;  
}

// this += s*x
void WeightVector::minus(WeightVector& x, double s) {
  my_snorm = 0.0; my_1norm=0.0;
  for (uint i=0; i<d; ++i) {
    my_v[i] *= my_a;
    my_v[i] -= (x[i] * s);
    my_snorm += my_v[i]*my_v[i];
    my_1norm +=fabs(my_v[i]);
  }
  my_a = 1.0;
    
}



// w =\min_u 1/2|u - w | + s |u|_1, returns the conjugate value of  max_u u^T w  - 1/2|u|^2_2 - s|u|_1
 double  WeightVector::proximal_l1(double s){
	 if(s==0.0)  return 0.0;
	 double conj_val=0.0;
	 my_snorm=0.0; my_1norm=0.0;
	 for(uint i=0; i<d; ++i){
		   my_v[i]*=my_a;
		   double abs_w=fabs(my_v[i]);
		   if(abs_w<s) my_v[i]=0.0;
		   else{
			  conj_val +=abs_w*abs_w/2;
			  my_v[i] *= (1 - s/abs_w);
		      conj_val -= s*(s/2 + fabs(my_v[i]));
		   }
		   my_snorm +=my_v[i]*my_v[i];
		   my_1norm +=fabs(my_v[i]);
	   }
	 my_a=1.0;
	 return conj_val;
  }


void WeightVector::print(std::ostream& os) {
  os<<d<<" "; //for reading using libsvmread
  for(uint i=0; i < d; ++i) {
    if (my_v[i] != 0.0)
      os << i << ":" << (my_v[i]*my_a) << " "; 
  }
  os << std::endl;
}

//--------------------------------------------------------------------------
double operator* (simple_sparse_vector& u, WeightVector& v) {
  double outcome = 0.0;
  for( simple_sparse_vector_iterator it = u.my_vec.begin(); 
      it != u.my_vec.end(); it++) {
    outcome += ((*it).second * v[(*it).first]);
  }
  return outcome;
}

//-----------------------------------------------------------------------------
double operator* (WeightVector& v, simple_sparse_vector& u) {
  return (u*v);
}


//--------------------------------------------------------------------------
double operator* (dense_vector& u, WeightVector& v) {
  double outcome = 0.0;
  uint d= v.dimension();
  if(u.dimension() !=d){
	      std::cerr << "multiplying WeightVector of size "
			<< d << " to a dense vector of size "
			<< u.dimension() << " is not permitted" << std::endl;
	      exit(EXIT_FAILURE);
  }
  for(uint i=0; i<d; i++) {
    outcome += u[i]*v[i];
  }
  return outcome;
}

//-----------------------------------------------------------------------------
double operator* (WeightVector& v, dense_vector& u) {
  return (u*v);
}

template double operator* (std::vector<double>& u, WeightVector& v);
template double operator* (std::vector<float>& u, WeightVector& v);

//--------------------------------------------------------------------------
template<typename T>
double operator* (std::vector<T>& u, WeightVector& v) {
  double outcome = 0.0;
  uint d= v.dimension();
  if((uint)u.size() !=d){
	      std::cerr << "multiplying WeightVector of size "
			<< d << " to a dense vector of size "
			<< u.size() << " is not permitted" << std::endl;
	      exit(EXIT_FAILURE);
  }
  for(uint i=0; i<d; i++) {
    outcome += u[i]*v[i];
  }
  return outcome;
}

//-----------------------------------------------------------------------------
template double operator* (WeightVector& v, std::vector<float>& u) ;

template double operator* (WeightVector& v, std::vector<double>& u) ;

template<typename T>
double operator* (WeightVector& v, std::vector<T>& u) {
  return (u*v);
}


double add_return_snorm (WeightVector& u, WeightVector& v) {
  double outcome = 0.0, sum;
  if(u.dimension()!=v.dimension()){
	  std::cerr << "dimension of the two vectors for adding does not match"<<std::endl;
	  exit(EXIT_FAILURE);
  }
  for(uint i=0; i< u.dimension(); i++ ){
	   sum= u[i]+ v[i];
	   outcome += sum*sum;
  }
  return outcome;
}

double add_return_snorm (WeightVector& u, WeightVector& v, double s) {
  double outcome = 0.0, sum;
  if(u.dimension()!=v.dimension()){
	  std::cerr << "dimension of the two vectors for adding does not match"<<std::endl;
	  exit(EXIT_FAILURE);
  }
  for(uint i=0; i< u.dimension(); i++ ){
	   sum= u[i]+ v[i]*s;
	   outcome += sum*sum;
  }
  return outcome;
}

double add_return_snorm (WeightVector& u, WeightVector& v, double s1, double s2) {
  double outcome = 0.0, sum;
  if(u.dimension()!=v.dimension()){
	  std::cerr << "dimension of the two vectors for adding does not match"<<std::endl;
	  exit(EXIT_FAILURE);
  }
  for(uint i=0; i< u.dimension(); i++ ){
	   sum= u[i]*s1 + v[i]*s2;
	   outcome += sum*sum;
  }
  return outcome;
}

long get_runtime(void)
{
  clock_t start;
  start = clock();
  return((long)((double)start/(double)CLOCKS_PER_SEC));
}
